// Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0

package com.amazonaws.iot.idt.model;

import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.immutables.value.Value;

import javax.annotation.Nullable;
import java.util.Map;

@IDTModel
@Value.Immutable
@JsonSerialize(as = GetContextValueRequest.class)
interface IDTGetContextStringRequest {
    // A get context string request.

    /**
     * The string to do placeholder evaluation on.
     *
     * Placeholder evaluation:
     * - Any placeholders of the form {{jsonPath}} within the string
     *   will be queried from the context and replaced with their actual values.
     *   For example, if the context was {"foo": "bar"}, then "a {{foo}} c" would result in "a bar c".
     * - A string without placeholders is returned unmodified.
     *
     * When getting anything but the root object, the leading "$." for JSON paths within placeholders is optional.
     */
    String inputString();

    /**
     * Optional extra definitions to insert into the IDT context for just this request.
     * These definitions are merged at the root of the context object.
     *
     * For example, if {"foo" : "bar"} is provided here,
     * then "foo" is now a valid JSON path to query in the IDT context even though IDT does not provide that path.
     * It will be resolved to "bar".
     *
     * Note:
     *  - Using keys IDT already provides (e.g. testData) should be avoided. If there is a key collision,
     *    values will resolve to the extra definition value and not IDT's value.
     *  - Keys must not start with "_IDT_" - these are reserved for internal use.
     */
    @Nullable
    Map<String, String> extraDefs();
}
